=== modified file 'bzrlib/changeset.py'
--- bzrlib/changeset.py	2005-08-05 21:06:19 +0000
+++ bzrlib/changeset.py	2005-08-17 08:32:56 +0000
@@ -43,55 +43,7 @@
     return newdict
 
 
-class PatchApply(object):
-    """Patch application as a kind of content change"""
-    def __init__(self, contents):
-        """Constructor.
-
-        :param contents: The text of the patch to apply
-        :type contents: str"""
-        self.contents = contents
-
-    def __eq__(self, other):
-        if not isinstance(other, PatchApply):
-            return False
-        elif self.contents != other.contents:
-            return False
-        else:
-            return True
-
-    def __ne__(self, other):
-        return not (self == other)
-
-    def apply(self, filename, conflict_handler, reverse=False):
-        """Applies the patch to the specified file.
-
-        :param filename: the file to apply the patch to
-        :type filename: str
-        :param reverse: If true, apply the patch in reverse
-        :type reverse: bool
-        """
-        input_name = filename+".orig"
-        try:
-            os.rename(filename, input_name)
-        except OSError, e:
-            if e.errno != errno.ENOENT:
-                raise
-            if conflict_handler.patch_target_missing(filename, self.contents)\
-                == "skip":
-                return
-            os.rename(filename, input_name)
-            
-
-        status = patch.patch(self.contents, input_name, filename, 
-                                    reverse)
-        os.chmod(filename, os.stat(input_name).st_mode)
-        if status == 0:
-            os.unlink(input_name)
-        elif status == 1:
-            conflict_handler.failed_hunks(filename)
-
-        
+       
 class ChangeUnixPermissions(object):
     """This is two-way change, suitable for file modification, creation,
     deletion"""
@@ -374,27 +326,30 @@
 
 
 class Diff3Merge(object):
-    def __init__(self, base_file, other_file):
-        self.base_file = base_file
-        self.other_file = other_file
+    def __init__(self, file_id, base, other):
+        self.file_id = file_id
+        self.base = base
+        self.other = other
 
     def __eq__(self, other):
         if not isinstance(other, Diff3Merge):
             return False
-        return (self.base_file == other.base_file and 
-                self.other_file == other.other_file)
+        return (self.base == other.base and 
+                self.other == other.other and self.file_id == other.file_id)
 
     def __ne__(self, other):
         return not (self == other)
 
     def apply(self, filename, conflict_handler, reverse=False):
-        new_file = filename+".new" 
+        new_file = filename+".new"
+        base_file = self.base.readonly_path(self.file_id)
+        other_file = self.other.readonly_path(self.file_id)
         if not reverse:
-            base = self.base_file
-            other = self.other_file
+            base = base_file
+            other = other_file
         else:
-            base = self.other_file
-            other = self.base_file
+            base = other_file
+            other = base_file
         status = patch.diff3(new_file, filename, base, other)
         if status == 0:
             os.chmod(new_file, os.stat(filename).st_mode)
@@ -402,7 +357,14 @@
             return
         else:
             assert(status == 1)
-            conflict_handler.merge_conflict(new_file, filename, base, other)
+            def get_lines(filename):
+                my_file = file(base, "rb")
+                lines = my_file.readlines()
+                my_file.close()
+            base_lines = get_lines(base)
+            other_lines = get_lines(other)
+            conflict_handler.merge_conflict(new_file, filename, base_lines, 
+                                            other_lines)
 
 
 def CreateDir():
@@ -1060,13 +1022,10 @@
     def rename_conflict(self, id, this_name, base_name, other_name):
         raise RenameConflict(id, this_name, base_name, other_name)
 
-    def move_conflict(self, id, inventory):
-        this_dir = inventory.this.get_dir(id)
-        base_dir = inventory.base.get_dir(id)
-        other_dir = inventory.other.get_dir(id)
+    def move_conflict(self, id, this_dir, base_dir, other_dir):
         raise MoveConflict(id, this_dir, base_dir, other_dir)
 
-    def merge_conflict(self, new_file, this_path, base_path, other_path):
+    def merge_conflict(self, new_file, this_path, base_lines, other_lines):
         os.unlink(new_file)
         raise MergeConflict(this_path)
 
@@ -1100,8 +1059,8 @@
     def missing_for_rename(self, filename):
         raise MissingForRename(filename)
 
-    def missing_for_merge(self, file_id, inventory):
-        raise MissingForMerge(inventory.other.get_path(file_id))
+    def missing_for_merge(self, file_id, other_path):
+        raise MissingForMerge(other_path)
 
     def new_contents_conflict(self, filename, other_contents):
         raise NewContentsConflict(filename)
@@ -1328,49 +1287,30 @@
         self.full_path = full_path
         self.stat_result = stat_result
 
-def generate_changeset(tree_a, tree_b, inventory_a=None, inventory_b=None):
-    return ChangesetGenerator(tree_a, tree_b, inventory_a, inventory_b)()
+def generate_changeset(tree_a, tree_b, interesting_ids=None):
+    return ChangesetGenerator(tree_a, tree_b, interesting_ids)()
 
 class ChangesetGenerator(object):
-    def __init__(self, tree_a, tree_b, inventory_a=None, inventory_b=None):
+    def __init__(self, tree_a, tree_b, interesting_ids=None):
         object.__init__(self)
         self.tree_a = tree_a
         self.tree_b = tree_b
-        if inventory_a is not None:
-            self.inventory_a = inventory_a
-        else:
-            self.inventory_a = tree_a.inventory()
-        if inventory_b is not None:
-            self.inventory_b = inventory_b
-        else:
-            self.inventory_b = tree_b.inventory()
-        self.r_inventory_a = self.reverse_inventory(self.inventory_a)
-        self.r_inventory_b = self.reverse_inventory(self.inventory_b)
+        self._interesting_ids = interesting_ids
 
-    def reverse_inventory(self, inventory):
-        r_inventory = {}
-        for entry in inventory.itervalues():
-            if entry.id is None:
-                continue
-            r_inventory[entry.id] = entry
-        return r_inventory
+    def iter_both_tree_ids(self):
+        for file_id in self.tree_a:
+            yield file_id
+        for file_id in self.tree_b:
+            if file_id not in self.tree_a:
+                yield file_id
 
     def __call__(self):
         cset = Changeset()
-        for entry in self.inventory_a.itervalues():
-            if entry.id is None:
-                continue
-            cs_entry = self.make_entry(entry.id)
+        for file_id in self.iter_both_tree_ids():
+            cs_entry = self.make_entry(file_id)
             if cs_entry is not None and not cs_entry.is_boring():
                 cset.add_entry(cs_entry)
 
-        for entry in self.inventory_b.itervalues():
-            if entry.id is None:
-                continue
-            if not self.r_inventory_a.has_key(entry.id):
-                cs_entry = self.make_entry(entry.id)
-                if cs_entry is not None and not cs_entry.is_boring():
-                    cset.add_entry(cs_entry)
         for entry in list(cset.entries.itervalues()):
             if entry.parent != entry.new_parent:
                 if not cset.entries.has_key(entry.parent) and\
@@ -1384,49 +1324,55 @@
                     cset.add_entry(parent_entry)
         return cset
 
-    def get_entry_parent(self, entry, inventory):
-        if entry is None:
-            return None
-        if entry.path == "./.":
-            return NULL_ID
-        dirname = os.path.dirname(entry.path)
-        if dirname == ".":
-            dirname = "./."
-        parent = inventory[dirname]
-        return parent.id
-
-    def get_path(self, entry, tree):
-        if entry is None:
-            return (None, None)
-        if entry.path == ".":
-            return ""
-        return entry.path
-
-    def make_basic_entry(self, id, only_interesting):
-        entry_a = self.r_inventory_a.get(id)
-        entry_b = self.r_inventory_b.get(id)
+    def iter_inventory(self, tree):
+        for file_id in tree:
+            yield self.get_entry(file_id, tree)
+
+    def get_entry(self, file_id, tree):
+        if file_id not in tree:
+            return None
+        return tree.tree.inventory[file_id]
+
+    def get_entry_parent(self, entry):
+        if entry is None:
+            return None
+        return entry.parent_id
+
+    def get_path(self, file_id, tree):
+        if not tree.has_id(file_id):
+            return None
+        path = tree.id2path(file_id)
+        if path == '':
+            return './.'
+        else:
+            return path
+
+    def make_basic_entry(self, file_id, only_interesting):
+        entry_a = self.get_entry(file_id, self.tree_a)
+        entry_b = self.get_entry(file_id, self.tree_b)
         if only_interesting and not self.is_interesting(entry_a, entry_b):
             return None
-        parent = self.get_entry_parent(entry_a, self.inventory_a)
-        path = self.get_path(entry_a, self.tree_a)
-        cs_entry = ChangesetEntry(id, parent, path)
-        new_parent = self.get_entry_parent(entry_b, self.inventory_b)
-
-
-        new_path = self.get_path(entry_b, self.tree_b)
+        parent = self.get_entry_parent(entry_a)
+        path = self.get_path(file_id, self.tree_a)
+        cs_entry = ChangesetEntry(file_id, parent, path)
+        new_parent = self.get_entry_parent(entry_b)
+
+        new_path = self.get_path(file_id, self.tree_b)
 
         cs_entry.new_path = new_path
         cs_entry.new_parent = new_parent
         return cs_entry
 
     def is_interesting(self, entry_a, entry_b):
+        if self._interesting_ids is None:
+            return True
         if entry_a is not None:
-            if entry_a.interesting:
-                return True
-        if entry_b is not None:
-            if entry_b.interesting:
-                return True
-        return False
+            file_id = entry_a.file_id
+        elif entry_b is not None:
+            file_id = entry_b.file_id
+        else:
+            return False
+        return file_id in self._interesting_ids
 
     def make_boring_entry(self, id):
         cs_entry = self.make_basic_entry(id, only_interesting=False)
@@ -1484,15 +1430,6 @@
             if stat_a.st_ino == stat_b.st_ino and \
                 stat_a.st_dev == stat_b.st_dev:
                 return None
-            if file(full_path_a, "rb").read() == \
-                file(full_path_b, "rb").read():
-                return None
-
-            patch_contents = patch.diff(full_path_a, 
-                                        file(full_path_b, "rb").read())
-            if patch_contents is None:
-                return None
-            return PatchApply(patch_contents)
 
         a_contents = self.get_contents(stat_a, full_path_a)
         b_contents = self.get_contents(stat_b, full_path_b)

=== modified file 'bzrlib/commands.py'
--- bzrlib/commands.py	2005-08-12 15:37:01 +0000
+++ bzrlib/commands.py	2005-08-17 08:32:56 +0000
@@ -1455,37 +1455,54 @@
 
 
 class cmd_merge(Command):
-    """Perform a three-way merge of trees.
-    
-    The SPEC parameters are working tree or revision specifiers.  Working trees
-    are specified using standard paths or urls.  No component of a directory
-    path may begin with '@'.
-    
-    Working tree examples: '.', '..', 'foo@', but NOT 'foo/@bar'
-
-    Revisions are specified using a dirname/@revno pair, where dirname is the
-    branch directory and revno is the revision within that branch.  If no revno
-    is specified, the latest revision is used.
-
-    Revision examples: './@127', 'foo/@', '../@1'
-
-    The OTHER_SPEC parameter is required.  If the BASE_SPEC parameter is
-    not supplied, the common ancestor of OTHER_SPEC the current branch is used
-    as the BASE.
-
+    """Perform a three-way merge.
+    
+    The branch is the branch you will merge from.  By default, it will merge
+    the latest revision.  If you specify a revision, that revision will be
+    merged.  If you specify two revisions, the first will be used as a BASE, 
+    and the second one as OTHER.  Revision numbers are always relative to the
+    specified branch.
+    
+    Examples:
+
+    To merge the latest revision from bzr.dev
+    bzr merge ../bzr.dev
+
+    To merge changes up to and including revision 82 from bzr.dev
+    bzr merge -r 82 ../bzr.dev
+
+    To merge the changes introduced by 82, without previous changes:
+    bzr merge -r 81..82 ../bzr.dev
+    
     merge refuses to run if there are any uncommitted changes, unless
     --force is given.
     """
-    takes_args = ['other_spec', 'base_spec?']
-    takes_options = ['force', 'merge-type']
+    takes_args = ['branch?']
+    takes_options = ['revision', 'force', 'merge-type']
 
-    def run(self, other_spec, base_spec=None, force=False, merge_type=None):
+    def run(self, branch='.', revision=None, force=False, 
+            merge_type=None):
         from bzrlib.merge import merge
         from bzrlib.merge_core import ApplyMerge3
         if merge_type is None:
             merge_type = ApplyMerge3
-        merge(parse_spec(other_spec), parse_spec(base_spec),
-              check_clean=(not force), merge_type=merge_type)
+
+        if revision is None or len(revision) < 1:
+            base = (None, None)
+            other = (branch, -1)
+        else:
+            if len(revision) == 1:
+                other = (branch, revision[0])
+                base = (None, None)
+            else:
+                assert len(revision) == 2
+                if None in revision:
+                    raise BzrCommandError(
+                        "Merge doesn't permit that revision specifier.")
+                base = (branch, revision[0])
+                other = (branch, revision[1])
+            
+        merge(other, base, check_clean=(not force), merge_type=merge_type)
 
 
 class cmd_revert(Command):

=== modified file 'bzrlib/merge.py'
--- bzrlib/merge.py	2005-08-05 21:06:19 +0000
+++ bzrlib/merge.py	2005-08-17 08:32:56 +0000
@@ -48,6 +48,15 @@
             d_file.write(line)
         os.chmod(dest, 0777 & os.stat(source).st_mode)
 
+    def dump(self, lines, dest):
+        """Copy the text and mode of a file
+        :param source: The path of the file to copy
+        :param dest: The distination file to create
+        """
+        d_file = file(dest, "wb")
+        for line in lines:
+            d_file.write(line)
+
     def add_suffix(self, name, suffix, last_new_name=None):
         """Rename a file to append a suffix.  If the new name exists, the
         suffix is added repeatedly until a non-existant name is found
@@ -72,7 +81,7 @@
         self.conflicts += 1
         
 
-    def merge_conflict(self, new_file, this_path, base_path, other_path):
+    def merge_conflict(self, new_file, this_path, base_lines, other_lines):
         """
         Handle diff3 conflicts by producing a .THIS, .BASE and .OTHER.  The
         main file will be a version with diff3 conflicts.
@@ -82,8 +91,8 @@
         :param other_path: Path to the file text for the OTHER tree
         """
         self.add_suffix(this_path, ".THIS")
-        self.copy(base_path, this_path+".BASE")
-        self.copy(other_path, this_path+".OTHER")
+        self.dump(base_lines, this_path+".BASE")
+        self.dump(other_lines, this_path+".OTHER")
         os.rename(new_file, this_path)
         self.conflict("Diff3 conflict encountered in %s" % this_path)
 
@@ -108,17 +117,6 @@
         if not self.ignore_zero:
             print "%d conflicts encountered.\n" % self.conflicts
             
-class SourceFile(object):
-    def __init__(self, path, id, present=None, isdir=None):
-        self.path = path
-        self.id = id
-        self.present = present
-        self.isdir = isdir
-        self.interesting = True
-
-    def __repr__(self):
-        return "SourceFile(%s, %s)" % (self.path, self.id)
-
 def get_tree(treespec, temp_root, label):
     location, revno = treespec
     branch = find_branch(location)
@@ -133,22 +131,9 @@
     return branch, MergeTree(base_tree, temp_path)
 
 
-def abspath(tree, file_id):
-    path = tree.inventory.id2path(file_id)
-    if path == "":
-        return "./."
-    return "./" + path
-
 def file_exists(tree, file_id):
     return tree.has_filename(tree.id2path(file_id))
     
-def inventory_map(tree):
-    inventory = {}
-    for file_id in tree.inventory:
-        path = abspath(tree, file_id)
-        inventory[path] = SourceFile(path, file_id)
-    return inventory
-
 
 class MergeTree(object):
     def __init__(self, tree, tempdir):
@@ -157,18 +142,29 @@
             self.root = tree.basedir
         else:
             self.root = None
-        self.inventory = inventory_map(tree)
         self.tree = tree
         self.tempdir = tempdir
         os.mkdir(os.path.join(self.tempdir, "texts"))
         self.cached = {}
 
+    def __iter__(self):
+        return self.tree.__iter__()
+
     def __contains__(self, file_id):
-        return id in self.tree
+        return file_id in self.tree
+
+    def get_file(self, file_id):
+        return self.tree.get_file(file_id)
 
     def get_file_sha1(self, id):
         return self.tree.get_file_sha1(id)
 
+    def id2path(self, file_id):
+        return self.tree.id2path(file_id)
+
+    def has_id(self, file_id):
+        return self.tree.has_id(file_id)
+
     def readonly_path(self, id):
         if id not in self.tree:
             return None
@@ -256,14 +252,11 @@
              source_file.interesting = source_file.id in interesting_ids
 
 
-def generate_cset_optimized(tree_a, tree_b, inventory_a, inventory_b,
-                            interesting_ids=None):
+def generate_cset_optimized(tree_a, tree_b, interesting_ids=None):
     """Generate a changeset.  If interesting_ids is supplied, only changes
     to those files will be shown.  Metadata changes are stripped.
     """ 
-    if interesting_ids is not None:
-        set_interesting(inventory_a, inventory_b, interesting_ids)
-    cset =  generate_changeset(tree_a, tree_b, inventory_a, inventory_b)
+    cset =  generate_changeset(tree_a, tree_b, interesting_ids)
     for entry in cset.entries.itervalues():
         entry.metadata_change = None
     return cset
@@ -273,26 +266,23 @@
                 ignore_zero=False, merge_type=ApplyMerge3, backup_files=False,
                 interesting_ids=None):
 
-    def merge_factory(base_file, other_file):
-        contents_change = merge_type(base_file, other_file)
+    def merge_factory(file_id, base, other):
+        contents_change = merge_type(file_id, base, other)
         if backup_files:
             contents_change = BackupBeforeChange(contents_change)
         return contents_change
-    
-    def generate_cset(tree_a, tree_b, inventory_a, inventory_b):
-        return generate_cset_optimized(tree_a, tree_b, inventory_a, inventory_b,
-                                       interesting_ids)
 
     this_tree = get_tree((this_branch.base, None), tempdir, "this")[1]
 
     def get_inventory(tree):
-        return tree.inventory
+        return tree.tree.inventory
 
     inv_changes = merge_flex(this_tree, base_tree, other_tree,
-                             generate_cset, get_inventory,
+                             generate_cset_optimized, get_inventory,
                              MergeConflictHandler(base_tree.root,
                                                   ignore_zero=ignore_zero),
-                             merge_factory=merge_factory)
+                             merge_factory=merge_factory, 
+                             interesting_ids=interesting_ids)
 
     adjust_ids = []
     for id, path in inv_changes.iteritems():
@@ -300,7 +290,7 @@
             if path == '.':
                 path = ''
             else:
-                assert path.startswith('./')
+                assert path.startswith('./'), "path is %s" % path
             path = path[2:]
         adjust_ids.append((path, id))
     if len(adjust_ids) > 0:
@@ -312,9 +302,24 @@
     old_entries = this_branch.read_working_inventory()
     new_inventory = {}
     by_path = {}
+    new_entries_map = {} 
+    for path, file_id in new_entries:
+        if path is None:
+            continue
+        new_entries_map[file_id] = path
+
+    def id2path(file_id):
+        path = new_entries_map.get(file_id)
+        if path is not None:
+            return path
+        entry = old_entries[file_id]
+        if entry.parent_id is None:
+            return entry.name
+        return os.path.join(id2path(entry.parent_id), entry.name)
+        
     for file_id in old_entries:
         entry = old_entries[file_id]
-        path = old_entries.id2path(file_id)
+        path = id2path(file_id)
         new_inventory[file_id] = (path, file_id, entry.parent_id, entry.kind)
         by_path[path] = file_id
     

=== modified file 'bzrlib/merge_core.py'
--- bzrlib/merge_core.py	2005-07-28 18:46:35 +0000
+++ bzrlib/merge_core.py	2005-08-17 08:32:56 +0000
@@ -6,15 +6,16 @@
 
 class ApplyMerge3:
     """Contents-change wrapper around merge3.Merge3"""
-    def __init__(self, base_file, other_file):
-        self.base_file = base_file
-        self.other_file = other_file
+    def __init__(self, file_id, base, other):
+        self.file_id = file_id
+        self.base = base
+        self.other = other
  
     def __eq__(self, other):
         if not isinstance(other, ApplyMerge3):
             return False
-        return (self.base_file == other.base_file and 
-                self.other_file == other.other_file)
+        return (self.base == other.base and 
+                self.other == other.other and self.file_id == other.file_id)
 
     def __ne__(self, other):
         return not (self == other)
@@ -23,14 +24,19 @@
     def apply(self, filename, conflict_handler, reverse=False):
         new_file = filename+".new" 
         if not reverse:
-            base = self.base_file
-            other = self.other_file
+            base = self.base
+            other = self.other
         else:
-            base = self.other_file
-            other = self.base_file
-        m3 = Merge3(file(base, "rb").readlines(), 
-                    file(filename, "rb").readlines(), 
-                    file(other, "rb").readlines())
+            base = self.other
+            other = self.base
+        def get_lines(tree):
+            if self.file_id not in tree:
+                raise Exception("%s not in tree" % self.file_id)
+                return ()
+            return tree.get_file(self.file_id).readlines()
+        base_lines = get_lines(base)
+        other_lines = get_lines(other)
+        m3 = Merge3(base_lines, file(filename, "rb").readlines(), other_lines)
 
         new_conflicts = False
         output_file = file(new_file, "wb")
@@ -48,7 +54,8 @@
             os.rename(new_file, filename)
             return
         else:
-            conflict_handler.merge_conflict(new_file, filename, base, other)
+            conflict_handler.merge_conflict(new_file, filename, base_lines,
+                                            other_lines)
 
 
 class BackupBeforeChange:
@@ -69,44 +76,35 @@
         self.contents_change.apply(filename, conflict_handler, reverse)
 
 
-class ThreewayInventory(object):
-    def __init__(self, this_inventory, base_inventory, other_inventory):
-        self.this = this_inventory
-        self.base = base_inventory
-        self.other = other_inventory
 def invert_invent(inventory):
     invert_invent = {}
-    for key, value in inventory.iteritems():
-        invert_invent[value.id] = key
+    for file_id in inventory:
+        path = inventory.id2path(file_id)
+        if path == '':
+            path = './.'
+        else:
+            path = './' + path
+        invert_invent[file_id] = path
     return invert_invent
 
-def make_inv(inventory):
-    return Inventory(invert_invent(inventory))
-        
 
 def merge_flex(this, base, other, changeset_function, inventory_function,
-               conflict_handler, merge_factory):
-    this_inventory = inventory_function(this)
-    base_inventory = inventory_function(base)
-    other_inventory = inventory_function(other)
-    inventory = ThreewayInventory(make_inv(this_inventory),
-                                  make_inv(base_inventory), 
-                                  make_inv(other_inventory))
-    cset = changeset_function(base, other, base_inventory, other_inventory)
-    new_cset = make_merge_changeset(cset, inventory, this, base, other, 
+               conflict_handler, merge_factory, interesting_ids):
+    cset = changeset_function(base, other, interesting_ids)
+    new_cset = make_merge_changeset(cset, this, base, other, 
                                     conflict_handler, merge_factory)
-    result = apply_changeset(new_cset, invert_invent(this_inventory),
+    result = apply_changeset(new_cset, invert_invent(this.tree.inventory),
                              this.root, conflict_handler, False)
     conflict_handler.finalize()
     return result
 
     
 
-def make_merge_changeset(cset, inventory, this, base, other, 
+def make_merge_changeset(cset, this, base, other, 
                          conflict_handler, merge_factory):
     new_cset = changeset.Changeset()
     def get_this_contents(id):
-        path = os.path.join(this.root, inventory.this.get_path(id))
+        path = this.readonly_path(id)
         if os.path.isdir(path):
             return changeset.dir_create
         else:
@@ -116,9 +114,10 @@
         if entry.is_boring():
             new_cset.add_entry(entry)
         else:
-            new_entry = make_merged_entry(entry, inventory, conflict_handler)
+            new_entry = make_merged_entry(entry, this, base, other, 
+                                          conflict_handler)
             new_contents = make_merged_contents(entry, this, base, other, 
-                                                inventory, conflict_handler, 
+                                                conflict_handler,
                                                 merge_factory)
             new_entry.contents_change = new_contents
             new_entry.metadata_change = make_merged_metadata(entry, base, other)
@@ -126,65 +125,76 @@
 
     return new_cset
 
-def make_merged_entry(entry, inventory, conflict_handler):
+class ThreeWayConflict(Exception):
+    def __init__(self, this, base, other):
+        self.this = this
+        self.base = base
+        self.other = other
+        msg = "Conflict merging %s %s and %s" % (this, base, other)
+        Exception.__init__(self, msg)
+
+def threeway_select(this, base, other):
+    """Returns a value selected by the three-way algorithm.
+    Raises ThreewayConflict if the algorithm yields a conflict"""
+    if base == other:
+        return this
+    elif base == this:
+        return other
+    elif other == this:
+        return this
+    else:
+        raise ThreeWayConflict(this, base, other)
+
+
+def make_merged_entry(entry, this, base, other, conflict_handler):
     from bzrlib.trace import mutter
-    this_name = inventory.this.get_name(entry.id)
-    this_parent = inventory.this.get_parent(entry.id)
-    this_dir = inventory.this.get_dir(entry.id)
-    if this_dir is None:
-        this_dir = ""
-
-    base_name = inventory.base.get_name(entry.id)
-    base_parent = inventory.base.get_parent(entry.id)
-    base_dir = inventory.base.get_dir(entry.id)
-    if base_dir is None:
-        base_dir = ""
-    other_name = inventory.other.get_name(entry.id)
-    other_parent = inventory.other.get_parent(entry.id)
-    other_dir = inventory.base.get_dir(entry.id)
-    if other_dir is None:
-        other_dir = ""
+    def entry_data(file_id, tree):
+        assert hasattr(tree, "__contains__"), "%s" % tree
+        if file_id not in tree:
+            return (None, None, "")
+        entry = tree.tree.inventory[file_id]
+        my_dir = tree.id2path(entry.parent_id)
+        if my_dir is None:
+            my_dir = ""
+        return entry.name, entry.parent_id, my_dir 
+    this_name, this_parent, this_dir = entry_data(entry.id, this)
+    base_name, base_parent, base_dir = entry_data(entry.id, base)
+    other_name, other_parent, other_dir = entry_data(entry.id, other)
     mutter("Dirs: this, base, other %r %r %r" % (this_dir, base_dir, other_dir))
     mutter("Names: this, base, other %r %r %r" % (this_name, base_name, other_name))
-    if base_name == other_name:
-        old_name = this_name
-        new_name = this_name
-    else:
-        if this_name != base_name and this_name != other_name:
-            conflict_handler.rename_conflict(entry.id, this_name, base_name,
-                                             other_name)
-        else:
-            old_name = this_name
-            new_name = other_name
-
-    if base_parent == other_parent:
-        old_parent = this_parent
-        new_parent = this_parent
-        old_dir = this_dir
-        new_dir = this_dir
-    else:
-        if this_parent != base_parent and this_parent != other_parent:
-            conflict_handler.move_conflict(entry.id, inventory)
-        else:
-            old_parent = this_parent
-            old_dir = this_dir
-            new_parent = other_parent
-            new_dir = other_dir
-    if old_name is not None and old_parent is not None:
-        old_path = os.path.join(old_dir, old_name)
-    else:
-        old_path = None
+    old_name = this_name
+    try:
+        new_name = threeway_select(this_name, base_name, other_name)
+    except ThreeWayConflict:
+        new_name = conflict_handler.rename_conflict(entry.id, this_name, 
+                                                    base_name, other_name)
+
+    old_parent = this_parent
+    try:
+        new_parent = threeway_select(this_parent, base_parent, other_parent)
+    except ThreeWayConflict:
+        new_parent = conflict_handler.move_conflict(entry.id, this_dir,
+                                                    base_dir, other_dir)
+    def get_path(name, parent):
+        if name is not None and parent is not None:
+            parent_dir = {this_parent: this_dir, other_parent: other_dir, 
+                          base_parent: base_dir}
+            directory = parent_dir[parent]
+            return os.path.join(directory, name)
+        else:
+            assert name is None and parent is None
+            return None
+
+    old_path = get_path(old_name, old_parent)
+        
     new_entry = changeset.ChangesetEntry(entry.id, old_parent, old_path)
-    if new_name is not None and new_parent is not None:
-        new_entry.new_path = os.path.join(new_dir, new_name)
-    else:
-        new_entry.new_path = None
+    new_entry.new_path = get_path(new_name, new_parent)
     new_entry.new_parent = new_parent
     mutter(repr(new_entry))
     return new_entry
 
 
-def make_merged_contents(entry, this, base, other, inventory, conflict_handler,
+def make_merged_contents(entry, this, base, other, conflict_handler,
                          merge_factory):
     contents = entry.contents_change
     if contents is None:
@@ -192,13 +202,10 @@
     this_path = this.readonly_path(entry.id)
     def make_merge():
         if this_path is None:
-            return conflict_handler.missing_for_merge(entry.id, inventory)
-        base_path = base.readonly_path(entry.id)
-        other_path = other.readonly_path(entry.id)    
-        return merge_factory(base_path, other_path)
+            return conflict_handler.missing_for_merge(entry.id, 
+                                                      other.id2path(entry.id))
+        return merge_factory(entry.id, base, other)
 
-    if isinstance(contents, changeset.PatchApply):
-        return make_merge()
     if isinstance(contents, changeset.ReplaceContents):
         if contents.old_contents is None and contents.new_contents is None:
             return None
@@ -211,7 +218,8 @@
             if this_path is None or not os.path.exists(this_path):
                 return contents
             else:
-                this_contents = file(this_path, "rb").read()
+                this_contents = changeset.FileCreate(file(this_path, 
+                                                     "rb").read())
                 if this_contents == contents.new_contents:
                     return None
                 else:
@@ -259,11 +267,41 @@
 import unittest
 import tempfile
 import shutil
+from bzrlib.inventory import InventoryEntry, RootEntry
+from osutils import file_kind
+class FalseTree(object):
+    def __init__(self, realtree):
+        self._realtree = realtree
+        self.inventory = self
+
+    def __getitem__(self, file_id):
+        entry = self.make_inventory_entry(file_id)
+        if entry is None:
+            raise KeyError(file_id)
+        return entry
+        
+    def make_inventory_entry(self, file_id):
+        path = self._realtree.inventory.get(file_id)
+        if path is None:
+            return None
+        if path == "":
+            return RootEntry(file_id)
+        dir, name = os.path.split(path)
+        kind = file_kind(self._realtree.abs_path(path))
+        for parent_id, path in self._realtree.inventory.iteritems():
+            if path == dir:
+                break
+        if path != dir:
+            raise Exception("Can't find parent for %s" % name)
+        return InventoryEntry(file_id, name, kind, parent_id)
+
+
 class MergeTree(object):
     def __init__(self, dir):
         self.dir = dir;
         os.mkdir(dir)
         self.inventory = {'0': ""}
+        self.tree = FalseTree(self)
     
     def child_path(self, parent, name):
         return os.path.join(self.inventory[parent], name)
@@ -301,6 +339,16 @@
     def readonly_path(self, id):
         return self.full_path(id)
 
+    def __contains__(self, file_id):
+        return file_id in self.inventory
+
+    def get_file(self, file_id):
+        path = self.readonly_path(file_id)
+        return file(path, "rb")
+
+    def id2path(self, file_id):
+        return self.inventory[file_id]
+
     def change_path(self, id, path):
         new = os.path.join(self.dir, self.inventory[id])
         os.rename(self.abs_path(self.inventory[id]), self.abs_path(path))
@@ -336,7 +384,17 @@
                 tree.remove_file(id)
             if other or base:
                 change = self.cset.entries[id].contents_change
-                assert isinstance(change, changeset.ReplaceContents)
+                if change is None:
+                    change = changeset.ReplaceContents(None, None)
+                    self.cset.entries[id].contents_change = change
+                    def create_file(tree):
+                        return changeset.FileCreate(tree.get_file(id).read())
+                    if not other:
+                        change.new_contents = create_file(self.other)
+                    if not base:
+                        change.old_contents = create_file(self.base)
+                else:
+                    assert isinstance(change, changeset.ReplaceContents)
                 if other:
                     change.new_contents=None
                 if base:
@@ -432,12 +490,9 @@
         os.chmod(tree.full_path(id), mode)
 
     def merge_changeset(self, merge_factory):
-        all_inventory = ThreewayInventory(Inventory(self.this.inventory),
-                                          Inventory(self.base.inventory), 
-                                          Inventory(self.other.inventory))
         conflict_handler = changeset.ExceptionConflictHandler(self.this.dir)
-        return make_merge_changeset(self.cset, all_inventory, self.this,
-                                    self.base, self.other, conflict_handler,
+        return make_merge_changeset(self.cset, self.this, self.base,
+                                    self.other, conflict_handler,
                                     merge_factory)
 
     def apply_inv_change(self, inventory_change, orig_inventory):
@@ -570,8 +625,8 @@
 
     def test_contents_merge3(self):
         """Test diff3 merging"""
-        def backup_merge(base_file, other_file):
-            return BackupBeforeChange(ApplyMerge3(base_file, other_file))
+        def backup_merge(file_id, base, other):
+            return BackupBeforeChange(ApplyMerge3(file_id, base, other))
         builder = self.contents_test_success(backup_merge)
         def backup_exists(file_id):
             return os.path.exists(builder.this.full_path(file_id)+"~")
@@ -594,6 +649,9 @@
         builder.add_file("2", "0", "name3", "text2", 0655)
         builder.change_contents("2", base="text5")
         builder.add_file("3", "0", "name5", "text3", 0744)
+        builder.add_file("4", "0", "name6", "text4", 0744)
+        builder.remove_file("4", base=True)
+        assert not builder.cset.entries["4"].is_boring()
         builder.change_contents("3", this="text6")
         cset = builder.merge_changeset(merge_factory)
         assert(cset.entries["1"].contents_change is not None)
@@ -603,6 +661,7 @@
             assert(isinstance(cset.entries["2"].contents_change,
                           merge_factory))
         assert(cset.entries["3"].is_boring())
+        assert(cset.entries["4"].is_boring())
         builder.apply_changeset(cset)
         assert(file(builder.this.full_path("1"), "rb").read() == "text4" )
         assert(file(builder.this.full_path("2"), "rb").read() == "text2" )

=== modified file 'bzrlib/patch.py'
--- bzrlib/patch.py	2005-07-26 15:25:58 +0000
+++ bzrlib/patch.py	2005-08-17 08:32:56 +0000
@@ -40,28 +40,6 @@
     return status 
 
 
-def diff(orig_file, mod_str, orig_label=None, mod_label=None):
-    """Compare two files, and produce a patch.
-
-    :param orig_file: path to the old file
-    :type orig_file: str
-    :param mod_str: Contents of the new file
-    :type mod_str: str
-    :param orig_label: The label to use for the old file
-    :type orig_label: str
-    :param mod_label: The label to use for the new file
-    :type mod_label: str
-    """
-    args = ["diff", "-u" ]
-    if orig_label is not None and mod_label is not None:
-        args.extend(("-L", orig_label, "-L", mod_label))
-    args.extend(("--", orig_file, "-"))
-    patch, stderr, status = write_to_cmd(args, mod_str)
-    if status == 0:
-        return None
-    else:
-        return patch
-
 def diff3(out_file, mine_path, older_path, yours_path):
     def add_label(args, label):
         args.extend(("-L", label))

=== modified file 'bzrlib/selftest/blackbox.py'
--- bzrlib/selftest/blackbox.py	2005-08-02 14:02:45 +0000
+++ bzrlib/selftest/blackbox.py	2005-08-17 08:32:56 +0000
@@ -361,8 +361,16 @@
         file('hello', 'wt').write('foo')
         self.runbzr('add hello')
         self.runbzr('commit -m setup hello')
+
+        file('goodbye', 'wt').write('baz')
+        self.runbzr('add goodbye')
+        self.runbzr('commit -m setup goodbye')
         
         file('hello', 'wt').write('bar')
+        file('goodbye', 'wt').write('qux')
         self.runbzr('revert hello')
         self.check_file_contents('hello', 'foo')
+        self.check_file_contents('goodbye', 'qux')
+        self.runbzr('revert')
+        self.check_file_contents('goodbye', 'baz')
 

=== modified file 'bzrlib/tree.py'
--- bzrlib/tree.py	2005-08-05 21:06:19 +0000
+++ bzrlib/tree.py	2005-08-17 08:32:56 +0000
@@ -161,6 +161,11 @@
     def __contains__(self, file_id):
         return file_id in self._inventory
 
+    def get_file_sha1(self, file_id):
+        assert self._inventory[file_id].kind == "root_directory"
+        return None
+
+
 
 
 ######################################################################

